#include "Constant.h"
#include "math.h"
#include <string>
#include <cstring>
#include <iostream>
#include "SingleVariantTest.h"
#include "Burden.h"
#include "snpHWE.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>

String SVT::geneMapfile = "refFlat_hg19.txt";
bool SVT::LRT = false;
bool SVT::SCORE = false;
bool SVT::WALD = false;
int SVT::Xstart = 2699520;
int SVT::Xend = 154931044;
int SVT::nFounder=0;
bool SVT::dosage =false;
bool SVT::genoFromPed =false;
bool SVT::genoFromVCF =false;
bool SVT::FounderFreq =false;
String SVT::xLabel = "X";
int SVT::maleLabel = 1;
int SVT::femaleLabel =2;
int SVT::seed = 1;
int SVT::simulate = 0;

bool SVT::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status = false;
   mac =0.0;
miss = 0.0;

   StringArray markerName;
   markerName.AddTokens(ped.markerNames[markerid],":");
   if(markerName.Length()!=2)
   {
      printf("Marker %s is not in chr:pos format, will skip.\n",ped.markerNames[markerid].c_str());
      return status;
   }
   chr = markerName[0];
   pos = atoi(markerName[1].c_str());
   founderMaf = markerMaf = 0.0;

   int alleleCount = CountAlleles(ped, markerid);
   if(alleleCount!=2)
      return status;

   String ref = ped.GetMarkerInfo(markerid)->alleleLabels[1];
   String alt = ped.GetMarkerInfo(markerid)->alleleLabels[2];

   bool XStatus = false;
   if(AutoFit::fitX)
   {
      int found = markerName[0].FastFind(xLabel,0);
      if(found!=-1)
      {
	 if(pos > Xstart && pos < Xend)
	 {
	    XStatus=true;
	 }
      }
   }
   int n=0;
nFounder=0;
   double fmaf = 0.0, maf=0.0;
   genotype.resize(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   //Loop through samples according to trans.pPheno 
   int idx=0;
   for(int f=0;f<ped.familyCount;f++)
   {
      for(int p=0;p<trans.pPheno[f].Length();p++)
      {
	 if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	 {
miss++;
	    idx++;
	    continue;
	 }
	 if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	 {
	    if(XStatus)
	    {
	       if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
	       {
		  genotype[idx] = 1.0;
	       }
	       if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
	       {
		  genotype[idx] = 2.0;
	       }
	    }
	    else
	    {
	       genotype[idx] = 1.0;
	    }
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	 {
	    if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	    {
	       genotype[idx] = 2.0;
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	    {
	       genotype[idx] = 0.0;
	    }
	 }
	 maf+=genotype[idx];
	 n++;
	 int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	 if(founder != -1)
	 {
	    fmaf += genotype[idx];
	    nFounder++;
	 }
	 idx++;
      }
   }

   maf /= 2.0*n;
   fmaf /= 2.0*nFounder;
   if(fmaf==0.0 && maf>0)
      fmaf = 1.0/(2.0*nFounder);

miss /= trans.persons;

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;

   bool flip=false;
   if(FounderFreq && fmaf>0.5)
   {
      flip=true;
   }
   else if(maf>0.5) 
   {
      flip=true;
   }

   if(flip)
   {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    genotype[i] = 2.0-genotype[i];
	 }
      }
      rareAllele = ref;
      refAllele = alt;
   }
   else
   {
      rareAllele = alt;
      refAllele = ref;
   }

   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 mac+=genotype[i];
      }
   }
   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }
   return status;
}

void SVT::Run(Pedigree & ped, FastTransform & trans,FastFit & engine, String & traitName,FILE * log)
{
   if(!LRT && !SCORE && !WALD) 
      return;

   int traitNum = engine.GetTraitID(ped,traitName.c_str(),log);
   FILE * SCOREoutput;
   FILE * LRToutput;
   FILE * WALDoutput;
   String filename;

   Eigen::VectorXf residuals;
   if(!FastFit::unrelated)
   {
      residuals.resize(trans.persons);
      Eigen::VectorXf beta;
      beta.resize(engine.betaHat.Length());
      for(int i=0;i<engine.betaHat.Length();i++)
      {
	 beta[i] = engine.betaHat[i];
      }
      residuals = trans.UDY - trans.UDX*beta;
   }
   else
   {
      if(simulate>0)
      {
	 residuals.resize(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    residuals[i] = engine.residuals_unrelated[i];
      }
   }
      String label;
label = "dummy";
      int markerId = ped.GetMarkerID(label);

   if(SCORE)
   {
      filename = Burden::outputFile + traitName + ".singlevar.score.txt";
      SCOREoutput = fopen(filename,"wt");
      PrintHeader(SCOREoutput,engine,trans);
   }
   if(LRT)
   {
      filename = Burden::outputFile + traitName + ".singlevar.LRT.txt";
      LRToutput = fopen(filename,"wt");
      PrintHeader(LRToutput,engine,trans);
   }
   if(WALD)
   {
      filename = Burden::outputFile + traitName + ".singlevar.wald.txt";
      WALDoutput = fopen(filename,"wt");
      PrintHeader(WALDoutput,engine,trans);
   }

   printf("\n\nStart single variant association tests for trait %s...\n",engine.traitName.c_str());
   Vector sigma2;
   Eigen::VectorXf transGeno; 
   transGeno.resize(trans.persons);
   sigma2.Dimension(trans.persons);
   double L0,L1,chisq,pvalue,effSize;

   int bad =0;
   if(LRT)
   {
      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
	 L0=engine.EvaluateMulti(trans,false);
      }
      else
      {
	 L0 = engine.Evaluate(engine.deltaHat,trans,false);
      }
   }
   int markerCount =0;

   if(dosage || genoFromVCF)
   {
      reader.open(GroupFromAnnotation::vcfInput,header);
      int numSamples = header.getNumSamples();
      while(reader.readRecord(record))
      {
	 markerCount++;
	 if(!GetGenotypeVector(trans,ped,markerCount-1))
	 {
	    bad++;
	    fprintf(log,"Variant %s:%d is monomorphic thus not analyzed.\n",chr.c_str(),pos);
	    continue;
	 }
	 if(FastFit::unrelated)
	 {
	    for(int i=0;i<trans.persons;i++)
	       sigma2[i] = engine.sigma2Hat;
	 }
	 else
	 {
	    if(!trans.empKin)
	    {
	       transGeno.setZero();
	       int index=0;
	       for (int f = 0; f < ped.familyCount; f++)
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   transGeno[index] += trans.UD(block+i,block+c)*genotype[block+c];
			}
			index++;
		     }
		  }
	    } else transGeno = trans.UD * genotype;
	 }
	 if(LRT || WALD)
	 {
	    FastTransform  fullTrans;
	    FastFit fullEngine;

	    fullTrans.persons = trans.persons;
	    fullTrans.fixedEffects = trans.fixedEffects +1;
	    if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	    {
	       int covariateId = ped.GetCovariateID("MarkerTested");

	       fullTrans.pPheno = new IntArray [ped.familyCount];
	       // If we are using covariates, we only consider individuals
	       // for which all covariates have been recorded
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  fullTrans.pPheno[f].Dimension(0);
		  for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		  {
		     if (ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
		     {
			fullTrans.pPheno[f].Push(i);
		     }
		  }
	       }

	       int index =0;
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (fullTrans.pPheno[f].Length() > 0)
		  {
		     int count = fullTrans.pPheno[f].Length();
		     for(int i=0;i<count;i++)
		     {
			ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotype[index];
			index++;
		     }
		  }
	       }
	       AutoFit fullAutoEngine;
	       fullAutoEngine.useCovariates = true;
	       fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
	       fullAutoEngine.fitX = AutoFit::fitX;

	       //Fit the alternative
	       fullAutoEngine.FitPolygenicModels(ped,traitNum,true);
	       effSize = fullAutoEngine.betaHat[fullAutoEngine.betaHat.Length()-1];

	       fullTrans.Y.Copy(trans.Y);
	       fullTrans.X.Copy(trans.X);
	       fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
	       fullTrans.UY.Dimension(fullTrans.persons);
	       fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.transU.resize(0,0);
	       fullTrans.D.Dimension(0);

	       for(int i=0;i<fullTrans.persons;i++)
	       {
		  fullTrans.X[i][trans.fixedEffects] = genotype[i];
	       }
	       fullTrans.traitVar = fullAutoEngine.variance;
	       fullTrans.TransformPedkinX(fullAutoEngine, ped);

	       if(LRT)
	       {
		  fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
		  L1=fullEngine.EvaluateMulti(fullTrans,false);

		  chisq = 2.0*(L1-L0);
		  //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
		  if(chisq<0)
		     chisq = -chisq;
		  //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       }
	       /*
		  if(WALD)
		  {
		  fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
		  fullTrans.UDY.resize(fullTrans.persons);
		  fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

		  fullTrans.FinalizeTransform(0.0,0.0);
		  double chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1);
		  pvalue = pchisq(chisq, 1,0,0);
	       //pvalue = chidist(chisq, 1);
	       fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       }
		*/
	    }
	    else
	    {
	       fullTrans.D.Copy(trans.D);
	       fullTrans.UY.Copy(trans.UY);
	       fullTrans.transU= trans.transU;
	       fullTrans.UD.resize(fullTrans.persons,fullTrans.persons);

	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.UX.Copy(trans.UX);
	       fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
	       for(int row=0;row<fullTrans.persons;row++)
	       {
		  fullTrans.UX[row][fullTrans.fixedEffects-1] = transGeno[row];
	       }
	       fullEngine.OptimizeDelta(0.0001,fullTrans);
	       effSize = fullEngine.beta[fullTrans.fixedEffects-1];
	       if(LRT)
	       {
		  L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);
		  chisq = 2.0*(L1-L0);
		  if(chisq<0)
		     chisq = -chisq;
		  //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       }

	       if(WALD)
	       {
		  fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
		  fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;
		  //   printf("sigmag2 is %g;sigmae2 is %g\n",fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		  fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
		  fullTrans.UDY.resize(fullTrans.persons);
		  fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

		  //fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		  double chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       }
	    }
	 }
	 if(SCORE)
	 {
	    double sigma_g2 = engine.sigma_g2Hat;
	    double sigma_e2 = engine.sigma_e2Hat;

	    if(!FastFit::unrelated)
	    {
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	    }
	    else
	    {
	       for(int i=0;i<trans.persons;i++)
		  sigma2[i] = engine.sigma2Hat;
	    }

	    double numerator = 0.0;
	    double denominator = 1.0e-10;
	    if(FastFit::unrelated)
	    {
	       for(int i=0;i<trans.persons;i++)
	       {
		  //troubleshooting
		  //printf("%g\n",engine.residuals_unrelated[i]);
		  numerator += engine.residuals_unrelated[i]*genotype[i];
		  denominator += genotype[i]*genotype[i];
	       }
	       if(denominator==0)
	       {
		  effSize = _NAN_;
		  pvalue = _NAN_;
		  return;
	       }
	       effSize = numerator/denominator;
	       denominator *= engine.sigma2Hat;
	       //TOTO:TS
	       //printf("sigma2 is: %g\n",engine.sigma2Hat);

	       chisq = numerator*numerator/denominator;
	       pvalue = pchisq(chisq, 1,0,0);
	    }
	    else
	    {
	       numerator = residuals.dot(transGeno);
	       denominator = transGeno.dot(transGeno);

	       chisq = numerator*numerator/denominator;
	       pvalue = pchisq(chisq, 1,0,0);
	       //This is the rare allele effect size
	       effSize = numerator/denominator;
	    }
	    fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\t",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.0-miss,effSize,pvalue);
            double permute_pvalue=_NAN_;
            if(simulate >0)
            {
               /*
                  FILE * out;
                  String filename = Burden::outputFile + traitName + "."+chr+"."+pos+".replication";
                  out = fopen(filename,"wt");
                  FILE * report;
                  filename = Burden::outputFile + traitName + "."+chr+"."+pos+".replication.report";
                  report = fopen(filename,"wt");
                */
               permute_pvalue = EvaluatePermutation(ped,trans,founderMaf,chisq,residuals,sigma2,markerId);
               fprintf(SCOREoutput,"%g\n",permute_pvalue);
            }
            else
               fprintf(SCOREoutput,"\n");
	 }
      }
   }

   if(genoFromPed)
   {
      markerCount = ped.markerCount;
      for(int m = 0;m<ped.markerCount;m++) 
      {
	 if(!GetGenotypeVector(trans,ped,m))  
	 {
	    bad++;
	    fprintf(log,"Variant %s:%d is monomorphic.\n",chr.c_str(),pos);
	    continue;
	 }

	 if(FastFit::unrelated)
	 {
	    for(int i=0;i<trans.persons;i++)
	       sigma2[i] = engine.sigma2Hat;
	 }
	 else
	 {
	    if(trans.pedKin) 
	    {
	       transGeno.setZero();
	       int index=0; 
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   //		   transGeno[index] += trans.transU(block+i,block+c)*genotype[block+c];
			   transGeno[index] += trans.UD(block+i,block+c)*genotype[block+c];
			}
			index++;
		     }
		  }
	       }
	    }
	    if(trans.empKin)
	    {
	       transGeno = trans.UD * genotype;
	    }
	 }
	 if(LRT || WALD)
	 {
	    FastTransform  fullTrans;
	    FastFit fullEngine;

	    fullTrans.persons = trans.persons;
	    fullTrans.fixedEffects = trans.fixedEffects +1;
	    if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	    {
	       int covariateId = ped.GetCovariateID("MarkerTested");

	       fullTrans.pPheno = new IntArray [ped.familyCount];
	       // If we are using covariates, we only consider individuals
	       // for which all covariates have been recorded
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  fullTrans.pPheno[f].Dimension(0);
		  for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		  {
		     if (ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
		     {
			fullTrans.pPheno[f].Push(i);
		     }
		  }
	       }

	       int index =0;
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (fullTrans.pPheno[f].Length() > 0)
		  {
		     int count = fullTrans.pPheno[f].Length();
		     for(int i=0;i<count;i++)
		     {
			ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotype[index];
			index++;
		     }
		  }
	       }

	       AutoFit fullAutoEngine; 
	       fullAutoEngine.useCovariates = true;
	       fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
	       fullAutoEngine.fitX = AutoFit::fitX;

	       //Fit the alternative
	       fullAutoEngine.FitPolygenicModels(ped,traitNum,true);
	       effSize = fullAutoEngine.betaHat[fullAutoEngine.betaHat.Length()-1]; 

	       fullTrans.Y.Copy(trans.Y);
	       fullTrans.X.Copy(trans.X);
	       fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
	       fullTrans.UY.Dimension(fullTrans.persons);
	       fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.transU.resize(0,0);
	       fullTrans.D.Dimension(0);

	       for(int i=0;i<fullTrans.persons;i++)
	       {
		  fullTrans.X[i][trans.fixedEffects] = genotype[i];
	       }
	       fullTrans.traitVar = fullAutoEngine.variance;
	       fullTrans.TransformPedkinX(fullAutoEngine, ped);

	       if(LRT)
	       {
		  fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
		  L1=fullEngine.EvaluateMulti(fullTrans,false);

		  chisq = 2.0*(L1-L0);
		  //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
		  if(chisq<0)
		     chisq = -chisq;
		  //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.-miss,effSize,pvalue);
	       }

	       if(WALD)
	       {
		  fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
		  fullTrans.UDY.resize(fullTrans.persons);
		  fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

		  //fullTrans.FinalizeTransform(0.0,0.0);
		  double chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.0-miss,effSize,pvalue);
	       }
	    }
	    else
	    {
	       fullTrans.D.Copy(trans.D);
	       fullTrans.UY.Copy(trans.UY);
	       fullTrans.transU = trans.transU;
	       fullTrans.UD.resize(fullTrans.persons,fullTrans.persons);

	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.UX.Copy(trans.UX);
	       fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
	       for(int row=0;row<fullTrans.persons;row++)
	       {
		  fullTrans.UX[row][fullTrans.fixedEffects-1] = transGeno[row];
	       }
	       fullEngine.OptimizeDelta(0.0001,fullTrans);
	       effSize = fullEngine.beta[fullTrans.fixedEffects-1];

	       if(LRT)
	       {
		  L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);
		  chisq = 2.0*(L1-L0);
		  if(chisq<0)
		     chisq = -chisq;
		  //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.0-miss,effSize,pvalue);
	       }

	       if(WALD)
	       {
		  fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
		  fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;
		  //   printf("sigmag2 is %g;sigmae2 is %g\n",fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		  fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
		  fullTrans.UDY.resize(fullTrans.persons);
		  fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

		  //fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		  double chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1); 
		  pvalue = pchisq(chisq, 1,0,0);
		  //pvalue = chidist(chisq, 1);
		  fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.0-miss,effSize,pvalue);
	       }
	    }
	 }

	 if(SCORE)
	 {
	    //	    printf("genotype vector is:\n");
	    //	    for(int i=0;i<genotype.Length();i++)
	    //	       printf("%g ",genotype[i]);
	    //	    printf("\n");
	    //printf("founderFreq is %d\n",FounderFreq);
	    double sigma_g2 = engine.sigma_g2Hat;
	    double sigma_e2 = engine.sigma_e2Hat;
	    if(FastFit::unrelated)
	    {
	       for(int i=0;i<trans.persons;i++)
		  sigma2[i] = engine.sigma2Hat;
	    }
	    else
	    {
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	    }
	    double numerator = 0.0;
	    double denominator = 1.0e-10; 
	    if(FastFit::unrelated)
	    {
	       for(int i=0;i<trans.persons;i++)
	       {
		  //troubleshooting
		  //printf("%g\n",engine.residuals_unrelated[i]);
		  numerator += engine.residuals_unrelated[i]*genotype[i];
		  denominator += genotype[i]*genotype[i];
	       }
	       if(denominator==0)
	       {
		  effSize = _NAN_;
		  pvalue = _NAN_;
		  return;
	       }
	       effSize = numerator/denominator;
	       denominator *= engine.sigma2Hat;
	       //TOTO:TS
	       //printf("sigma2 is: %g\n",engine.sigma2Hat);

	       chisq = numerator*numerator/denominator;
	       pvalue = pchisq(chisq, 1,0,0);
	    }
	    else
	    {
	       numerator = residuals.dot(transGeno);
	       denominator = transGeno.dot(transGeno);

	       if(denominator==0)
	       {
		  fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str());
		  return;
	       }
	       else chisq = numerator*numerator/denominator;
	       //	    printf("numerator and denominator are: %g %g\n",numerator, denominator);
	       pvalue = pchisq(chisq, 1,0,0);
	       //This is the rare allele effect size
	       effSize = numerator/denominator;
	    }
	    fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%7.3f\t%g\t",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),1.0-miss,effSize,pvalue);

	    double permute_pvalue=_NAN_;
	    if(simulate >0)
	    {
	       /*
		  FILE * out;
		  String filename = Burden::outputFile + traitName + "."+chr+"."+pos+".replication";
		  out = fopen(filename,"wt");
		  FILE * report;
		  filename = Burden::outputFile + traitName + "."+chr+"."+pos+".replication.report";
		  report = fopen(filename,"wt");
		*/
	       permute_pvalue = EvaluatePermutation(ped,trans,founderMaf,chisq,residuals,sigma2,markerId);
	       fprintf(SCOREoutput,"%g\n",permute_pvalue);
	    }
	    else
	       fprintf(SCOREoutput,"\n");
	 }
      }
   }
   if(LRT)
   {
      fprintf(LRToutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
   } 
   if(SCORE)
   {
      fprintf(SCOREoutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
   }
   if(WALD)
   {
      fprintf(WALDoutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
   }
   if(SCORE)
      fclose(SCOREoutput);
   if(LRT)
      fclose(LRToutput);
   if(WALD)
      fclose(WALDoutput);

   //Restoring the pedigree covariate settings.
   if((AutoFit::fitX || AutoFit::fitSharedEnvironment) && (LRT || WALD))
   {
      ped.covariateNames.Delete(ped.covariateNames.Length()-1);
      ped.covariateLookup.Delete("MarkerTested");
      ped.covariateCount--;
      for(int i=0;i<ped.count;i++)
	 ped[i].covariates[ped.covariateCount] = 0.0;
   }
}

double SVT::EvaluatePermutation(Pedigree & ped, FastTransform & trans, double & founderFreq, double & chisq, Eigen::VectorXf & residuals, Vector & sigma2, int & m)
{
   double pvalue = _NAN_;
   //Save missing genotype patter
   Eigen::VectorXf transGeno_;
   Eigen::VectorXf genotype_;
   genotype_.resize(trans.persons);
   double numerator,denominator;
   Vector chisq_;
/*
   int * missing =  new int[trans.persons];
   int idx=0;
   for(int f=0;f<ped.familyCount;f++)
   {
      for(int p=0;p<trans.pPheno[f].Length();p++)
      {
	 if(ped[trans.pPheno[f][p]].isGenotyped(m))
	    missing[idx]=0;
	 else
	    missing[idx]=1;
	 idx++;
      }
   }
  */
 Random rand;

   //Simulate genotypes
   for(int i=0;i<simulate;i++)
   {
      //simulate genotypes (phased)
      for(int f=0;f<ped.familyCount;f++)
      {
	 Family * family = ped.families[f];
	 //allocate founder haplotypes
	 for (int j = 0; j < family->founders; j++)
	 {
	    ped[family->path[j]].markers[m][0]= (rand.Uniform(0.0,1.0)<founderMaf ? 2 : 1);
	    ped[family->path[j]].markers[m][1]= (rand.Uniform(0.0,1.0)<founderMaf ? 2 : 1);
	 }
	 //allocate children haplotyps
	 for (int j = 0; j < family->nonFounders; j++)
	 {
	    int index  = j + family->founders;
	    int person = family->path[index];
	    Person & p = ped[person];

	    int fid = p.father->traverse;
	    int mid = p.mother->traverse;
	    if(rand.Uniform(0.0,1.0)<0.5)
	       p.markers[m][0] = (family->ped[family->path[fid]]).markers[m].one;
	    else
	       p.markers[m][0] = (family->ped[family->path[fid]]).markers[m].two;

	    if(rand.Uniform(0.0,1.0)<0.5)
	       p.markers[m][1] = (family->ped[family->path[mid]]).markers[m].one;
	    else
	       p.markers[m][1] = (family->ped[family->path[mid]]).markers[m].two;
	 }
      }

      //fill in genotype vector
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
/*
	    if(missing[idx]==1)
	    {
	       genotype_[idx]=-1.0;
	       idx++;
	       continue;
	    }
*/
	    if(ped[trans.pPheno[f][p]].markers[m].isHeterozygous())
	    {
	       /*
		  if(XStatus)
		  {
		  if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
		  {
		  genotype_[idx] = 1.0;
		  }
		  if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		  {
		  genotype_[idx] = 2.0;
		  }
		  }
		  else
		*/
	       genotype_[idx] = 1.0;
	    }
	    else if(ped[trans.pPheno[f][p]].markers[m].isHomozygous())
	    {
	       if(ped[trans.pPheno[f][p]].markers[m].hasAllele(2))
	       {
		  genotype_[idx] = 2.0;
	       }
	       else if(ped[trans.pPheno[f][p]].markers[m].hasAllele(1))
	       {
		  genotype_[idx] = 0.0;
	       }
	    }
	    idx++;
	 }
      }
      double mac_=0.0;
      int n_=0;
      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype_[i]!=-1.0)
	 {
	    mac_+=genotype_[i];
	    n_++;
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype_[i]==-1.0)
	 {
	    genotype_[i] = 0.0;
	 }
	 else
	 {
	    genotype_[i] -= mac/n_;
	 }
      }
      //calculate association statistics
      if(FastFit::unrelated)
      {
	 numerator = residuals.dot(genotype_);
	 denominator = genotype_.dot(genotype_);
      }
      else
      {
	 transGeno_.resize(trans.persons);
	 if(!trans.empKin)
	 {
	    transGeno_.setZero();
	    int index=0;
	    for (int f = 0; f < ped.familyCount; f++)
	       if (trans.pPheno[f].Length() > 0)
	       {
		  int count = trans.pPheno[f].Length();
		  int block = index;
		  for(int i=0;i<count;i++)
		  {
		     for(int c=0;c<count;c++)
		     {
			transGeno_[index] += trans.UD(block+i,block+c)*genotype_[block+c];
		     }
		     index++;
		  }
	       }
	 } else transGeno_ = trans.UD * genotype_;
	 numerator = residuals.dot(transGeno_);
	 denominator = transGeno_.dot(transGeno_);
      }
      chisq_.Push(numerator*numerator/denominator);
      //write to output
   }

   //organize results
   chisq_.Push(chisq);
   chisq_.Sort();
   for(int i=0;i<chisq_.Length();i++)
   {
      if(chisq_[i]==chisq)
      {
	 pvalue = (double) (simulate+1-i)/simulate;
	 return pvalue;
      }
   }
   /*
      Vector marks;
      marks.Push(0.01);
      marks.Push(0.001);
      marks.Push(0.0001);
      marks.Push(0.00001);
      marks.Push(0.000001);
      marks.Push(0.0000001);
      marks.Push(0.00000001);
      for(int i=0;i<marks.Length();i++)
      {
      int idx = (int) (marks[i]*permutation);
      if(idx<3)
      {
      printf("%g\tNA\n",marks[i]);
      fprintf(report,"%g\tNA\n",marks[i]);
      }
      else
      {
      if(chisq_[idx] > chisq)
      {
      printf("%g\t0\n",marks[i]);
      fprintf(report,"%g\t0\n",marks[i]);
      }
      else 
      {
      printf("%g\t1\n",marks[i]);
      fprintf(report,"%g\t1\n",marks[i]);
      }
      }
      }
      fprintf(out,"chisq:%g\n",chisq);
      for(int i =0;i<permutation;i++)
      {
      fprintf(out,"%g\n",chisq_[i]);
   //printf("%g\n",chisq_[i]);
   }
    */
}

void SVT::PrintHeader(FILE * output,FastFit & engine,FastTransform & trans)
{
   double sigma_g2 = engine.sigma_g2Hat;
   double h2 = 100.0*sigma_g2/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv =100.0*engine.sigma_s2Hat/trans.traitVar;
   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, Xheritability = %6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, SharedEnv = %g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, Xheritability = %g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }
   fprintf(output,"There are %d individuals included in the analysis.\n",trans.persons);
   fprintf(output,"===================================================================\n");
   fprintf(output,"\nCHR\tPOS\tFOUNDER_AF\tALL_AF\tINFORMATIVE_AC\tREF_ALLELE\tALT_ALLELE\tCALL_RATE\tALT_ALLELE_EFFSIZE\tPVALUE\t");
   if(simulate > 0)
   {
      fprintf(output,"PERMUTATION_PVALUE\n");
   }
   else
      fprintf(output,"\n");
}

bool SVT::GetGenotypeVectorVCFDosage(FastTransform & trans, Pedigree & ped)
{
   founderMaf = markerMaf = 0.0;
   bool status = false;
   String alt = record.getAltStr();
   String ref = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   int n=0,nf=0;
   double fmaf = 0.0, maf=0.0;
   mac=0.0;
   miss = 0.0;

   VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

   bool XStatus = false;
   if(AutoFit::fitX)
   {  
      if(chr==xLabel && pos > Xstart && pos < Xend)
      {  
	 XStatus=true;
      }  
   }

   genotype.resize(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   //Loop through samples according to trans.pPheno 
   //read in dosage into genotype vector
   int idx=0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid;

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    //if dosage is not missing
	    if(strcmp((*geno).c_str(),".")>0)
	    {
	       double dose = atof((*geno).c_str());
	       if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==maleLabel)
	       {
		  genotype[idx] = 2;
	       }
	       else
		  genotype[idx] = dose;
	       maf += genotype[idx];
	       n++;

	       //judge if this is a founder
	       int founder = trans.foundersHash.Integer(sample);
	       if(founder != -1)
	       {
		  fmaf += genotype[idx];
		  nf++;
	       }
	    }
	    else 
	       miss++;
	 }
	 idx++;
      }
   }
   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0.)
      fmaf = 1.0/(2.0*nf);

   miss /= trans.persons;

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;

   /*
      bool flip=false;
      if(FounderFreq && fmaf>0.5)
      {
      flip=true;
      }
      else if(maf>0.5)
      {
      flip=true;
      }
      if(flip)
      {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
      if(genotype[i]!=-1.0)
      {
      genotype[i] = 2.0-genotype[i];
      }
      }
      rareAllele = ref;
      refAllele = alt;
      }
      else
      {
    */
   rareAllele = alt;
   refAllele = ref;
   //   }
   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 mac+=genotype[i];
      }
   }

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }

   return status;
}

bool SVT::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped)
{
   founderMaf = markerMaf = 0.0;
   bool status = false;

   String alt = record.getAltStr();
   String ref= record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   mac=0.0;
   miss = 0.0;
   double fmaf=0.0,maf=0.0;
   int n=0,nf=0;

   bool XStatus = false;
   if(AutoFit::fitX)
   {
      if(chr==xLabel && pos > Xstart && pos < Xend)
      {
	 XStatus=true;
      }
   }

   //Fill in genotype vector
   genotype.resize(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i] = -1.0;

   int idx=0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid.c_str();

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    int numGTs = record.getNumGTs(s);
	    Vector allele;

	    for(int jr= 0; jr < numGTs; jr++)
	    {
	       allele.Push(record.getGT(s,jr));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    //if no missing allele exists
	    if(bad==0)
	    {
	       int founder = trans.foundersHash.Integer(sample);

	       int sum=0;
	       for(int i=0;i<allele.Length();i++)
		  sum+=allele[i];

	       if(XStatus && sum >0 && ped[trans.pPheno[f][j]].sex==maleLabel)
	       {
		  genotype[idx] = 2;
	       }
	       else
		  genotype[idx] = sum;
	       maf += genotype[idx];
	       n++;
	       if(founder!=-1)
	       {
		  nf++;
		  fmaf+=genotype[idx];
	       }
	    }
	 }
	 idx++;
      }
   }
   //Adjust genotype here
   maf /= 2.0*n;
   fmaf /=2.0*nf;

   if(fmaf==0.0 && maf>0.0)
      fmaf = 1.0/(2.0*nf);

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;
   /*
      bool flip=false;
      if(FounderFreq && fmaf>0.5)
      {
      flip=true;
      }
      else if(maf>0.5)
      {
      flip=true;
      }

      if(flip)
      {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
      if(genotype[i]!=-1.0)
      {
      genotype[i] = 2.0-genotype[i];
      }
      }
      rareAllele = ref;
      refAllele = alt;
      }
      else
      {
    */
   rareAllele = alt;
   refAllele = ref;
   //   }

   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 miss++;
	 mac+=genotype[i];
      }
   }
   miss /=trans.persons;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }
   return status;
}

bool SVT::GetGenotypeVector(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status;
   if(dosage)
   {
      status = GetGenotypeVectorVCFDosage(trans,ped);
   }
   if(genoFromPed)
   {
      status = GetGenotypeVectorPED(trans,ped,markerid);
   }
   if(genoFromVCF && !dosage)
   {
      status = GetGenotypeVectorVCF(trans,ped);
   }
   return status;
}


